/************************************************************************************
  USBMASCONLN.ino   LN Version
   This sketch can be used by connecting ZUIKI's SWitch mascon via the USB HOST shield.
   The source is created by referring to the sample USB HID Boystick of the USB HOST shield.

   This sketch requires the following libraries:
    Loconet Library 1.1.4 
    USB_Host_Shield_2.0
    I2C 128x64 OLED Driver Library   (used the libraries included in DS HACX R4)

   2021/10/20 Rev03
   Copyright (c) 2021 Nagoden  All right reserved.
*************************************************************************************/
#include <usbhid.h>
#include <hiduniversal.h>
#include <usbhub.h>
#include <EEPROM.h>

// Satisfy IDE, which only needs to see the include statment in the ino.
#ifdef dobogusinclude
#include <spi4teensy3.h>
#endif
#include <SPI.h>
#include "hidjoystickrptparser.h"

//OLED library
#include "OzOLED.h"

//----------------------------------------------------------------------
//マスコン加減速定数
//ブレーキは減速数　加速は加速数
//0から1023までで速度制御するのでノッチの位置でどれだけ足し引きするかの定数　
#define CONST_EB 20
#define CONST_B8 15
#define CONST_B7 10
#define CONST_B6 8
#define CONST_B5 5
#define CONST_B4 4
#define CONST_B3 3
#define CONST_B2 2
#define CONST_B1 1
#define CONST_NN 1
#define CONST_P1 10
#define CONST_P2 5
#define CONST_P3 5
#define CONST_P4 5
#define CONST_P5 5
//マスコン速度リミット％
//Pノッチの位置でどこまでを最高とするかの率を％で指定
#define LIM_P1 10
#define LIM_P2 30
#define LIM_P3 50
#define LIM_P4 80
#define LIM_P5 100
//ボタンのファンクション割り当て(F0～F8までしかできません)
//99を指定するとそのボタンのファンクション機能は無効
#define BUTTON_A 2
#define BUTTON_B 3
#define BUTTON_Y 4
#define BUTTON_X 6
#define BUTTON_ZL 5
#define BUTTON_L 0
#define BUTTON_M 1
#define BUTTON_P 7
#define BUTTON_R 8

//----------------------------------------------------------------------
// Settings used in Loconet
#include <LocoNet.h>
#define ID 0x20                   //Throtte　ID　複数使う場合はここを調整することが必要
#define THROWN 0
#define CLOSED 32
#define PWR_LED_PIN A1           // PWR_LED(Power LED) on/off pin
#define RUN_LED_PIN A0           // RUN_LED(Connection LED) on/off pin
lnMsg        *LnPacket;
LnBuf        LnTxBuffer;

/*********************************************************************************
 Variable definition
 ********************************************************************************/
//走行用
int Direction = 0;  //0でFWD  1でRWD
long gSpeed = 0;    //SPEED MAX1023
int SpeedRate = 0;  //スピード比率（バー表示のための0-255の値）　
int gMspeed = 0;    //スピード値保持用     

//OLED表示用文字変数
char gText[16 + 1];
int gLocoAddresses[] = {0, 0, 0, 0};
int gbkLocoAddresses[] = {0, 0, 0, 0};  //アドレス変更時のバックアップ
int gswAddresses[] = {0, 0, 0, 0};
int gbkswAddresses[] = {0, 0, 0, 0};  //アクセサリスイッチアドレス変更時のバックアップ
char* space14 = "              ";

//DCCアドレス
int gdecLocoadddress = 0003 ;   //10進表記でのアドレス
uint16_t glocoadress = 0;       //バイナリでのアドレス
int Cmd_data = 0;               //コマンド用変数
//マスコンボタンのイベント発生時の変数
byte buttonBufRz = 0;
byte buttonBufX1 = 0;
byte buttonBufX2 = 0;
byte buttonBufY1 = 0;

//loconet用
uint8_t SlotNo = 0;           //コマンドステーションが管理するスロット番号で、機種によって扱える
                              //MAX数に制限があります。
int ValueFunction[28];        //LocoNetライブラリではF8までしか使えません。
int gTimeState = 0;           //コマンド受信待ちの処理のステータスフラグ
int gThotteError = 0;         //アドレスコマンド送出後のエラー取得用
int gStepmode = 3;            // 0:TH_SP_ST_28  1:TH_SP_ST_28_TRI  2:TH_SP_ST_28_ADV  
                              // 3:TH_SP_ST_128 4:TH_SP_ST_128_ADV 5:TH_SP_ST_14
boolean Power_Flg = false;    //線路電源　ON/OFF　フラグ

/*********************************************************************************
 Menu definition
 ********************************************************************************/
uint8_t SetAddress_mode;
int wXcursor;                 //メニュー制御ｘアドレスワークカーソル
int wadXcursor;               //メニュー制御ｘアクセサリワークカーソル

/*********************************************************************************
    Task Schedule
 ********************************************************************************/
unsigned long gPreviousL1 = 0; //   200ms interval  Acceleration / deceleration processing after throttle event
unsigned long gPreviousL2 = 0; //   500ms interval  Automatic deceleration in neutral
unsigned long gPreviousL3 = 0; //   400ms interval  Loconet Recive commmand

/********************************************************************************
   JoystickEventsクラスを継承
 ********************************************************************************/
class MyJoystickEvents : public JoystickEvents {
  public:
    void OnGamePadChanged(const GamePadEventData *evt);
};

USB Usb;
USBHub Hub(&Usb);
HIDUniversal Hid(&Usb);
MyJoystickEvents MyJoyEvents;            //Make it a new inherited class.
JoystickReportParser Joy(&MyJoyEvents);  //Use an instance of a new class

/*********************************************************************************
  Loconet Class定義

 ********************************************************************************/
LocoNetThrottleClass Throttle;
LocoNetClass Loconet;
uint32_t  LastThrottleTimerTick;
uint32_t  LastAdrTimerTick;
/*********************************************************************************
  Definition of the function
 ********************************************************************************/
/* LCD functions */
OzOLED LCD;

/*********************************************************************************
  SETUP

 ********************************************************************************/
void setup() {
  Serial.begin(115200);
#if !defined(__MIPSEL__)
  while (!Serial); // Wait for serial port to connect - used on Leonardo, Teensy and other boards with built-in USB CDC serial connection
#endif
  Serial.println("Start");

  if (Usb.Init() == -1)
    Serial.println("OSC did not start.");

  delay(200);

  if (!Hid.SetReportParser(0, &Joy))
    ErrorMessage<uint8_t > (PSTR("SetReportParser"), 1);


  /*First initialize the LocoNet interface*/ 
      LocoNet.init(7);
      Throttle.init(0, 0, ID);  //init(uint8_t UserData, uint8_t Options, uint16_t ThrottleId )
  
  /* OLED init */
      LCD.OutputDammy = outputDammyPulse;
      LCD.init();
      LCD.clearDisplay();           //clear the screen and set start position to top left corner
      LCD.setNormalDisplay();       //Set display to Normal mode
      LCD.setPageMode();            //Set addressing mode to Page Mode
      LCD.printString("USB Mascon", 0, 2);
      LCD.printString("Ver1.03", 0, 3);
      LCD.printString("Nagoden", 0, 5);
      delay(2000);

  //LED初期化
      pinMode(PWR_LED_PIN , OUTPUT);  // Power LED
      pinMode(RUN_LED_PIN , OUTPUT);  // Connect LED
      digitalWrite(PWR_LED_PIN, LOW);
      digitalWrite(RUN_LED_PIN, LOW);


  //EEPROMからアドレス読み出し

  
      EEPROM.get(0x000, glocoadress);
      if ((glocoadress > 0x270F) || (glocoadress == 0 ) ) {
          //EEPROMの値が9999より大きいか０の場合はアドレスを初期値３にする
          glocoadress = 3 ;
      }
      //Serial.print("EEPROM_R:");
      //Serial.println(glocoadress);
      gdecLocoadddress = glocoadress ;

  //表示用配列にセット
      disassemblyLocoaddress();

  //OLED初期表示
      clearDisplay() ;
      locoadddisp();
      dispMessageFirst();
  //方向表示
      Direction = 0;   // FWD
      dispDirection( Direction );
  //スピードバーの表示
      drawSpeedBar ( 9, 3, 0X00, 0);
  //画面
      SetAddress_mode = 0;

  //Reset task
      gPreviousL1 = millis();
      gPreviousL2 = millis();
      gPreviousL3 = millis();
}

/*********************************************************************************
  Loop処理
 ********************************************************************************/

void loop() {

  Usb.Task();

 /*********************************************
  * Loconet受信処理
   接続のインターフェースで設定が異なる
  *********************************************/

  // Check for any received LocoNet packets
  LnPacket = LocoNet.receive() ;
  if (LnPacket)  {
      //Serial.println("LnPacket");
        if (!LocoNet.processSwitchSensorMessage(LnPacket))
          Throttle.processMessage(LnPacket) ;
  }

  if (isTime(&LastThrottleTimerTick, 100)) {      // Locoには100ms周期でアクセスが必要だそうです。
      Throttle.process100msActions();
      if (Throttle.getState() != TH_ST_IN_USE) {    //no Slot - Blink LED F0
          digitalWrite(RUN_LED_PIN, LOW);
      } else
          digitalWrite(RUN_LED_PIN, HIGH);
      if (Throttle.getState() == TH_ST_SLOT_MOVE) {   //Blink off
          digitalWrite(RUN_LED_PIN, LOW);
      }
  }


  /***********************************
    タスクスケジュール処理
    マスコンの位置で加減速処理
  ***********************************/

  if ( (millis() - gPreviousL1) >= 300)  {
      //Reset task
      gPreviousL1 = millis();

      /***********************************
        マスコンの位置で加減速処理
        加減速の調整はこの中のgSpeedの足し引き
        している値をdefine値で調整可能
      ***********************************/
      //マスコンハンドルの文字表示
      switch (buttonBufRz)
      {
        case 0:  //EB
          gSpeed = gSpeed - CONST_EB ;
          break;
        case 0x05:  //B8
          gSpeed = gSpeed - CONST_B8 ;
          break;
        case 0x13:  //B7
          gSpeed = gSpeed - CONST_B7 ;
          break;
        case 0x20:  //B6
          gSpeed = gSpeed - CONST_B6 ;
          break;
        case 0x2E:  //B5
          gSpeed = gSpeed - CONST_B5 ;
          break;
        case 0x3C:  //B4
          gSpeed = gSpeed - CONST_B4 ;
          break;
        case 0x49:  //B3
          gSpeed = gSpeed - CONST_B3 ;
          break;
        case 0x57:  //B2
          gSpeed = gSpeed - CONST_B2 ;
          break;
        case 0x65:  //B1
          gSpeed = gSpeed - CONST_B1 ;
          break;
        case 0x80:  //N
          break;
        case 0x9F:  //P1
          if ((( gSpeed * 100) / 1023 ) < LIM_P1 ) {
            gSpeed = gSpeed + CONST_P1 ;                  //最初の加速は値が小さいうちは起動しないので加速を上げている
          }
          break;
        case 0xB7:  //P2
          if ((( gSpeed * 100) / 1023 ) < LIM_P2 ) { 
            gSpeed = gSpeed + CONST_P2 ;
          }
          break;
        case 0xCE:  //P3
          if ((( gSpeed * 100) / 1023 ) < LIM_P3 ) { 
            gSpeed = gSpeed + CONST_P3;
          }
          break;
        case 0xE6:  //P4
          if ((( gSpeed * 100) / 1023 ) < LIM_P4 ) {
            gSpeed = gSpeed + CONST_P4;
          }
          break;
        case 0xFF:  //P5
          if ((( gSpeed * 100) / 1023 ) <= LIM_P5 ) {
            gSpeed = gSpeed + CONST_P5;
          }
          break;
      }

      if ( gSpeed < 0  ) {
        gSpeed = 4 ;        //マスコンでの速度が０より小さい場合はスピードコマンド値の１になるように４を設定（下で４で割っているので）
      }
      if ( gSpeed > 1023 ) {
        gSpeed = 1023;
      }
      SpeedRate =  gSpeed / 4;
      //スピードバーの表示
      drawSpeedBar ( 9, 3, SpeedRate, 0);
      sprintf(gText, " %3d%c  ", ((word)( SpeedRate * 100) / 255) , 0x25 );
      printStringNOT ( gText , 9, 4);
      //スピードコマンド出力
      sendSpeed(SpeedRate);
  }

  /***********************************
    マスコンの位置がNの場合の減速処理
  ***********************************/
  if ( (millis() - gPreviousL2) >= 500)
  {
    //Reset task
      gPreviousL2 = millis();
    //マスコンハンドルがNの時で走行状態の場合に少しずつ減速
    if ((buttonBufRz == 0x80) && (gSpeed > 0) ) {
      gSpeed = gSpeed - CONST_NN ;
      drawSpeedBar ( 9, 3, SpeedRate, 0);   //スピードバーの表示
      sprintf(gText, " %3d%c  ", ((word)( SpeedRate * 100) / 255) , 0x25 );
      printStringNOT ( gText , 9, 4);
      sendSpeed(SpeedRate);                 //スピードコマンド出力
    }
  }
  /***********************************
    コマンド送出時の戻り値を確認するため一定
    の時間loop処理を回してコマンドの返りを
    待つタイムステート
  ***********************************/
  if ( (millis() - gPreviousL3) >= 600)
  {
    //Reset task
    //gPreviousL3 = millis();
    gTimeState = 1; 
  }


  /***********************************
    アイドル状態の時のキー処理
  ***********************************/
  //SetAddress_mode = 0 運転モード（何もボタン処理がない）
  //SetAddress_mode = 1 ZRボタンが押されアドレス変更モード
  //SetAddress_mode = 2 アドレス設定後のAボタンでDispatch処理する時にStealするか判定する状態
  //SetAddress_mode = 3 Steal判定後にStealしない場合のアドレス登録モード
  //SetAddress_mode = 4 Steal判定後にStealする場合のアドレス登録モード
  //SetAddress_mode = 5 アクセサリ（ポイント）用のコマンド送出モード
  
  if (SetAddress_mode == 0 ) {
    //ZRボタンでアドレス設定画面に変移
    if ((buttonBufX1 & 0x80) == 0x80 ) {    //ZRボタン
      buttonBufX1 = 0x00;
      SetAddress_mode = 1 ;                 //アドレス変更モードに設定
      dispMessageAddset();                  //アドレス設定用のメッセージ
      dispMessageMode(1);                   //モード表示　AD
      //カーソルの反転
      wXcursor = 0;
      printStringNOT("  ", wXcursor , 4);   //アドレス番号の下にカーソル表示
      //現在のアドレス保持
      for (int i = 0; i < 4; i++) {
        gbkLocoAddresses[i] = gLocoAddresses[i];
      }
    }
    /***********************************
      方向反転の処理

    ***********************************/
    if ((buttonBufX2) == 0x02 ) {      //→ボタン
      buttonBufX2 = 0x0f;
      delay(100);
      Direction = 0;   // FWD
      dispDirection( Direction );
    }
    if ((buttonBufX2) == 0x06 ) {      //←ボタン
      buttonBufX2 = 0x0f;
      delay(100);
      Direction = 1;   // RWD
      dispDirection( Direction );
    }
    /***********************************
      ファンクションキーの処理

    ***********************************/
    setFunction() ;

  }

  /***********************************
    アドレス設定処理
    　  *SetAddress_mode == 1の時のキー処理
  ***********************************/
  if (SetAddress_mode == 1 ) {
      if ((buttonBufX1 & 0x02) == 0x02 ) {      //Bボタン
          buttonBufX1 = 0x00;
          dispMessageClear() ;                  //メッセージクリア
          dispMessageMode(0);                   //モード表示　クリア
          delay(100);
          dispMessageFirst();                   //初期メッセージ表示
          SetAddress_mode = 0 ;                 //運転モード
          cursorClear();                        //カーソルの反転クリア
          //アドレス保持の復帰
          for (int i = 0; i < 4; i++) {
              gLocoAddresses[i] = gbkLocoAddresses[i];
              sprintf(gText, "%1d", gLocoAddresses[i]);
              printStringBig(gText , i * 2, 0);
          }
      }
      if ((buttonBufX1 & 0x04) == 0x04 ) {      //Aボタン
          buttonBufX1 = 0x00;
          dispMessageClear() ;                  //メッセージクリア
          SetAddress_mode = 0 ;                 //運転モード
          //アドレスを変数にセット(10進数)
          int w;
          w = ( gLocoAddresses[0] * 1000 ) + ( gLocoAddresses[1] * 100 ) + ( gLocoAddresses[2] * 10) + gLocoAddresses[3];
          //            Serial.println( w);
          glocoadress = w;
          gdecLocoadddress = w;
         /*********************************************
             ここでアドレスをコマンドステーションに送る処理
             とアドレスを記憶する処理を記述(ディスパッチ処理）
             接続のインターフェースで設定が異なる
          *********************************************/
           Throttle.idleAddress();
           Throttle.dispatchAddress(gdecLocoadddress);
           SetAddress_mode = 2;
           gPreviousL3 = millis();
           gTimeState = 0;
      }

      if ((buttonBufX1 & 0x08) == 0x08 ) {    //ボタンX →アドレス解放
          if (Throttle.getState() == TH_ST_IN_USE) {
              Throttle.freeAddress();                      
              delay(200);
              dispMessageClear();
              dispMessageFirst();
              dispMessageMode(0);              //モード表示　クリア
              dispDirection( Direction );  
              SetAddress_mode = 0;
          }
      }
      
      //アドレス設定モードでさらにDRが押された時
      if ((buttonBufX1 & 0x80) == 0x80 ) {    //ZRボタン
          buttonBufX1 = 0x00;
          dispMessageswAddset();              //アクセサリ用メッセージ表示
          dispMessageMode(2);                 //モード表示　SW
          swadddisp();                        //アクセサリ用アドレス表示
          SetAddress_mode = 5 ;               //クセサリ（ポイント）用のコマンド送出モード
          //カーソルの反転
          wadXcursor = 0;
          printStringNOT("  ", wadXcursor , 4);
          //現在のアドレス保持
          for (int i = 0; i < 4; i++) {
            gbkswAddresses[i] = gswAddresses[i];
          }
   
      }
  }
    /************************************************　
      アドレス設定のキー処理
      矢印キーとA・Bボタンでアドレス決定
      ←→キーで桁のカーソル移動
      ↑↓キーで数値のUp/Down
      Aボタンで決定　Bボタンでキャンセル
    *************************************************/
    if ((buttonBufX2 == 0x02) && (SetAddress_mode == 1 )) {      //→ボタン
        buttonBufX2 = 0x0f;
        delay(100);
        printString("  ", wXcursor * 2 , 4);
        wXcursor = wXcursor + 1;
        if ( wXcursor == 4 ) {
          wXcursor = 0;
        }
        printStringNOT("  ",  wXcursor * 2 , 4);
    }
    if ((buttonBufX2 == 0x06) && (SetAddress_mode == 1 )) {      //←ボタン
        buttonBufX2 = 0x0f;
        delay(100);
        printString("  ", wXcursor * 2 , 4);
        wXcursor = wXcursor - 1;
        if ( wXcursor < 0 ) {
          wXcursor = 3;
        }
        printStringNOT("  ",  wXcursor * 2 , 4);
    }
    if ((buttonBufX2 == 0x00) && (SetAddress_mode == 1 )) {      //↑ボタン
        buttonBufX2 = 0x0f;
        delay(100);
        gLocoAddresses[ wXcursor ] =  gLocoAddresses[ wXcursor ] + 1;
        if (gLocoAddresses[ wXcursor ] >= 10 ) {
           gLocoAddresses[ wXcursor ] = 0;
        }
        sprintf(gText, "%1d", gLocoAddresses[wXcursor]);
        printStringBig(gText , wXcursor * 2 , 0);
    }
    if ((buttonBufX2 == 0x04) && (SetAddress_mode == 1 )) {      //↓ボタン
        buttonBufX2 = 0x0f;
        delay(100);
        if (gLocoAddresses[ wXcursor ] == 0 ) {
          gLocoAddresses[ wXcursor ] = 9;
        }
        else {
          gLocoAddresses[ wXcursor ] =  gLocoAddresses[ wXcursor ] - 1;
        }
        sprintf(gText, "%1d", gLocoAddresses[wXcursor]);
        printStringBig(gText , wXcursor * 2 , 0);
    }

  /***********************************
    Dispatchしたあとのステータス処理
    　  *SetAddress_mode == 2の時のキー処理
  ***********************************/
    if ((SetAddress_mode == 2 )&& (gTimeState == 1)) {
         if ((gThotteError == 1)) {
            //他のスロットで同じアドレスを掴んでいるのでスチールするか確認
            SetAddress_mode = 4;
            dispMessageSteal();
         }
         else {
            //他のスロットで同じアドレスを掴んでいない場合はアドレス追加処理
            SetAddress_mode = 3;
            gPreviousL3 = millis();
            gTimeState = 0;
         }
    }

  /***********************************
    Dispatchしたあとのステータス処理
    　  *SetAddress_mode == 3の時のsetAdress処理
  ***********************************/
    if ((SetAddress_mode == 3 )&& (gTimeState == 1)) {
       Throttle.setAddress(gdecLocoadddress);   //アドレス追加処理
       cursorClear();                           //カーソルの反転クリア
       dispMessageClear();
       dispMessageFirst();
       dispMessageMode(0);                      //モード表示　クリア
       AddressEEPROM_Write();
       delay(100);
       SendStepmode(gStepmode);                 //ステップモードの設定
       Direction = Throttle.getDirection();
       dispDirection( Direction );  
       SetAddress_mode = 0;
    }  
  /*******************************************
    すでにスロットに同じ番号がある場合のスチール処理
    　  *SetAddress_mode == 2の時のキー処理
  ********************************************/
    if (SetAddress_mode == 4 ) {
        if ((buttonBufX1 & 0x02) == 0x02 ) {      //Bボタン
          buttonBufX1 = 0x00;
          dispMessageAddset();
          SetAddress_mode = 1;
        }
        if ((buttonBufX1 & 0x04) == 0x04 ) {      //Aボタン
          buttonBufX1 = 0x00;
          SetAddress_mode = 0;
          /*********************************************
          接続のインターフェースで設定が異なる
          *********************************************/
          Throttle.stealAddress(gdecLocoadddress);
          dispMessageClear();
          cursorClear();                           //カーソルの反転クリア
          dispMessageMode(0);                      //モード表示　クリア
          dispMessageFirst();
          AddressEEPROM_Write();
          SendStepmode(gStepmode);                 //ステップモードの設定
          delay(200);
          Direction = Throttle.getDirection();
          dispDirection( Direction );  
          /********************************************/
        }
    }

  /*******************************************
    アクセサリアドレス設定
    　  *SetAddress_mode == 5の時のキー処理
  ********************************************/
    if (SetAddress_mode == 5 ) {
        if ((buttonBufX1 & 0x02) == 0x02 ) {      //Bボタン
          buttonBufX1 = 0x00;
          locoadddisp();
          cursorClear();    //カーソルの反転クリア
          dispMessageMode(0);
          dispMessageAddset();
          dispMessageFirst();
          SetAddress_mode = 0;
        }
        if ((buttonBufX1 & 0x04) == 0x04 ) {      //Aボタン（CLOSED）
          buttonBufX1 = 0x00;
          //アドレスを変数にセット(10進数)
          int w;
          w = ( gswAddresses[0] * 1000 ) + ( gswAddresses[1] * 100 ) + ( gswAddresses[2] * 10) + gswAddresses[3];
          glocoadress = w;
          Loconet.requestSwitch(w,1,CLOSED);
        }
        if ((buttonBufX1 & 0x01) == 0x01 ) {      //Yボタン（THROWN）
          buttonBufX1 = 0x00;
          //アドレスを変数にセット(10進数)
          int w;
          w = ( gswAddresses[0] * 1000 ) + ( gswAddresses[1] * 100 ) + ( gswAddresses[2] * 10) + gswAddresses[3];
          glocoadress = w;
          Loconet.requestSwitch(w,1,THROWN);
        }
        if ((buttonBufX2) == 0x02 ) {             //→ボタン
            buttonBufX2 = 0x0f;
            delay(100);
            printString("  ", wadXcursor * 2 , 4);
            wadXcursor = wadXcursor + 1;
            if ( wadXcursor == 4 ) {
                wadXcursor = 0;
            }
            printStringNOT("  ",  wadXcursor * 2 , 4);
        }
        if ((buttonBufX2) == 0x06 ) {             //←ボタン
            buttonBufX2 = 0x0f;
            delay(100);
            printString("  ", wadXcursor * 2 , 4);
            wadXcursor = wadXcursor - 1;
            if ( wadXcursor < 0 ) {
                wadXcursor = 3;
            }
            printStringNOT("  ",  wadXcursor * 2 , 4);
        }
        if ((buttonBufX2 == 0x00 )&& (SetAddress_mode == 5 )) {      //↑ボタン
            buttonBufX2 = 0x0f;
            delay(100);
            gswAddresses[ wadXcursor ] =  gswAddresses[ wadXcursor ] + 1;
            if (gswAddresses[ wadXcursor ] >= 10 ) {
              gswAddresses[ wadXcursor ] = 0;
            }
            sprintf(gText, "%1d", gswAddresses[wadXcursor]);
            printStringBig(gText , wadXcursor * 2 , 0);
        }
        if ((buttonBufX2 == 0x04) && (SetAddress_mode == 5 )) {      //↓ボタン
            buttonBufX2 = 0x0f;
            delay(100);
            if (gswAddresses[ wadXcursor ] == 0 ) {
                gswAddresses[ wadXcursor ] = 9;
            }
            else {
              gswAddresses[ wadXcursor ] =  gswAddresses[ wadXcursor ] - 1;
            }
            sprintf(gText, "%1d", gswAddresses[wadXcursor]);
            printStringBig(gText , wadXcursor * 2 , 0);
        }
        
    }
}

/***********************************************************************
  各種関数
************************************************************************/
//EEPROMへのアドレス登録
void AddressEEPROM_Write (void) {
    EEPROM.put(0x000, glocoadress);
//  Serial.print("EEPROM_W:");
//  Serial.println(glocoadress);
}

// アドレスデータを４つの桁を表示用の配列にセット
void disassemblyLocoaddress(void){
    int work_Address = 0;
    int w;
    work_Address = gdecLocoadddress;
    gLocoAddresses[0] = work_Address / 1000 ;
    work_Address = work_Address - ((gLocoAddresses[0]) * 1000);
    gLocoAddresses[1] = work_Address / 100 ;
    work_Address = work_Address - (gLocoAddresses[1] * 100);
    gLocoAddresses[2] = work_Address / 10 ;
    work_Address = work_Address - (gLocoAddresses[2] * 10);
    gLocoAddresses[3] = work_Address;
}

//表示用の配列データをアドレスデータに変換
void assemblyLocoaddress(void){
    int work_Address;
    work_Address = ( gLocoAddresses[0] * 1000 ) + ( gLocoAddresses[1] * 100 ) + ( gLocoAddresses[2] * 10) + gLocoAddresses[3];
    //  Serial.println(work_Address);
    gdecLocoadddress = work_Address ;
}
//locoアドレス初期表示
void locoadddisp(void){
    //アドレス4桁表示
      for (int i = 0; i < 4; i++) {
        sprintf(gText, "%1d", gLocoAddresses[i]);
      printStringBig(gText , i * 2, 0);
      }
}
//アクセサリSWアドレス初期表示
void swadddisp(void){
    //アドレス3桁表示
      for (int i = 0; i < 4; i++) {
        sprintf(gText, "%1d", gswAddresses[i]);
      printStringBig(gText , (i * 2), 0);
      }
}

//進行方向表示　
void dispDirection(int n ) {
    Throttle.setDirection(n);
    if (n == 0) {
      //FWD表示
      printString ( "  FWD  " , 9, 2);
      drawIcon (14, 2, 2, 0);
      drawIcon (15, 2, 2, 0);
    }
    else {
      //RWD表示
      printString ( "  RWD  " , 9, 2);
      drawIcon (9, 2, 3, 0);
      drawIcon (10, 2, 3, 0);
    }
}

//OLEDのメッセージ枠のクリア
void dispMessageClear(void) {
    printString(space14 , 0, 5);
    printString(space14 , 0, 6);
    printString(space14 , 0, 7);
}
//OLEDのメッセージ 初期画面
void dispMessageFirst(void) {
    printString(space14 , 0, 5);
    printString("ZR:Set LocoAdr", 0, 6);
    printString("<>:Direction  ", 0, 7);
}
//OLEDのメッセージ アドレス設定画面
void dispMessageAddset(void) {
    printString("<>UpDn: Select", 0, 5);
    printString("A: Decision   ", 0, 6);
    printString("B: Exit       ", 0, 7);
}
//OLEDのメッセージ スイッチアドレス設定画面
void dispMessageswAddset(void) {
    printString("<>UpDn: Select", 0, 5);
    printString("A-> c  Y-> t  ", 0, 6);
    printString("B: Exit    ", 0, 7);
}
//OLEDのメッセージ アドレス設定　Steal確認画面
void dispMessageSteal(void) {
    printString("Steal y/n? ", 0, 5);
    printString("A: Yes     ", 0, 6);
    printString("B: No      ", 0, 7);
}
//OLEDのメッセージ モード画面
void dispMessageMode(int dmode) {
    switch(dmode){
      case 0:      //clear
          LCD.printString("  ", 9, 0);
          LCD.printString("  ", 9, 1);
          break; 
      case 1:      //LocoAddress
          LCD.printStringNOT("AD", 9, 0);
          LCD.printString("  ", 9, 1);
          break; 
      case 2:      //AccessoryAddress
          LCD.printString("  ", 9, 0);
          LCD.printStringNOT("SW", 9, 1);
          break; 
    }
}
//OLEDのメッセージ アドレス表示バーのカーソルクリア
void cursorClear(void){
      printString("        ", 0, 4);           //カーソルの反転クリア
}
//スロットルスピードコマンド処理
void sendSpeed(int spd) {
    //減速停止時にスピードが０にならないための対応処理
    if ((gMspeed > 0 ) && ( spd == 0 )){
        spd = 1 ;
        gMspeed = 0;
    }
    else {
        spd = spd / 2;   //本来ならスピードステップに合わせて計算すべき
        gMspeed = spd; //スピード値の保持
    }

    //コマンド送出
    if (spd == 0) {
      //spd=0は緊急停止　なので　ここでは送らない
      //本来１が緊急停止であるが０で緊急停止になっている
      //ライブラリで自動で緊急停止を送らないように制御されている
    }
    else {
      //スピードコマンド出力(loconet)
        Throttle.setSpeed( spd);
    }
}

//スピードステップの設定
void SendStepmode( int Stepmode){
    switch(Stepmode){
      case 0:        //0:TH_SP_ST_28
        Throttle.setSpeedSteps(TH_SP_ST_28);      // 28 step/ 3 BYTE PKT regular mode
        break;
      case 1:        //1:TH_SP_ST_28_TRI
        Throttle.setSpeedSteps(TH_SP_ST_28_TRI);  //28 step. Generate Trinary packets for this Mobile ADR
        break;
      case 2:        //2:TH_SP_ST_28_ADV
        Throttle.setSpeedSteps(TH_SP_ST_28_ADV);  //28 Step decoder ,Allow Advanced DCC consisting
        break;
      case 3:        //3:TH_SP_ST_128
        Throttle.setSpeedSteps(TH_SP_ST_128);     //128 speed mode packets
        break;
      case 4:        //4:TH_SP_ST_128_ADV
        Throttle.setSpeedSteps(TH_SP_ST_128_ADV); //128 Step decoder, Allow Advanced DCC consisting
        break;
      case 5:        //5:TH_SP_ST_14
        Throttle.setSpeedSteps(TH_SP_ST_14);      //14 step MODE
        break;
    }
}
/*****************************************************************************
//マスコンの押しボタンでファンクション送信
//ファンクションと押しボタンとのマッピングはここで変更してね
*****************************************************************************/
void setFunction(void) {
  //マスコンボタン処理その１
  if ((buttonBufX1 & 0x04) == 0x04 ) {    //ボタンA
    if (BUTTON_A != 99) {
        Throttle.setFunction( BUTTON_A , !ValueFunction[BUTTON_A]);        //ボタンA
    }    
  }
  if ((buttonBufX1 & 0x02) == 0x02 ) {    //ボタンB
    if (BUTTON_B != 99) {
        Throttle.setFunction( BUTTON_B , !ValueFunction[BUTTON_B]);        //ボタンB
    }
  }
  if ((buttonBufX1 & 0x01) == 0x01 ) {    //ボタンY
    if (BUTTON_Y != 99) {
        Throttle.setFunction( BUTTON_Y , !ValueFunction[BUTTON_Y]);        //ボタンY
    }    
  }
  if ((buttonBufX1 & 0x08) == 0x08 ) {    //ボタンX
    if (BUTTON_X != 99) {
        Throttle.setFunction( BUTTON_X , !ValueFunction[BUTTON_X]);        //ボタンX
    }    
  }
  if ((buttonBufX1 & 0x20) == 0x20 ) {    //ボタンR
    if (BUTTON_R != 99) {
        Throttle.setFunction( BUTTON_R , !ValueFunction[BUTTON_R]);        //ボタンX
    }    
  }
  if ((buttonBufX1 & 0x10) == 0x10 ) {    //ボタンL
    if (BUTTON_L != 99) {
        Throttle.setFunction( BUTTON_L , !ValueFunction[BUTTON_L]);        //ボタンL
    }    
  }
  if ((buttonBufX1 & 0x40) == 0x40 ) {    //ボタンZL
    if (BUTTON_ZL != 99) {
        Throttle.setFunction( BUTTON_ZL , !ValueFunction[BUTTON_ZL]);      //ボタンZL
    }    
    //マスコンのEBを入れるとZLボタンが同時にセットされる仕様であることに注意
  }
  if ((buttonBufY1 & 0x20) == 0x20 ) {                                     //ボタン●
      //線路電源ON/OFF
       lnMsg SendPacket ; 
       Power_Flg = !Power_Flg;
       if (Power_Flg == 1){
          //command送信
          SendPacket.data[ 0 ] = 0x83; 
          SendPacket.data[ 1 ] = 0x7C; 
          LocoNet.send(&SendPacket);
          digitalWrite(PWR_LED_PIN, HIGH);
       } else {
          //command送信
          SendPacket.data[ 0 ] = 0x82; 
          SendPacket.data[ 1 ] = 0x7D; 
          LocoNet.send(&SendPacket);
          digitalWrite(PWR_LED_PIN, LOW);
       }
  }
  if ((buttonBufY1 & 0x10) == 0x10 ) {                                    //ボタンHOME
      //Homeボタンは緊急停止（EBの機能とは違ってＤＣＣでの緊急停止です）
        gSpeed = 0;
        Throttle.setSpeed(gSpeed);     //緊急停止Loconet
  }
  if ((buttonBufY1 & 0x02) == 0x02 ) {                                    //ボタン+
    if (BUTTON_P != 99) {
        Throttle.setFunction( BUTTON_P , !ValueFunction[BUTTON_P]);       
    }    
  }
  if ((buttonBufY1 & 0x01) == 0x01 ) {                                    //ボタン-
    if (BUTTON_M != 99) {
        Throttle.setFunction( BUTTON_M , !ValueFunction[BUTTON_M]);       
    }    
  }
  buttonBufX1 = 0x00;
  buttonBufY1 = 0x00;

  //      ZRはアドレス設定用(使用禁止)
  //      if ((buttonBufX1 & 0x80) == 0x80 ) {    //ボタンZR
  //      }
}
//--------------------------------------------------------------------------------------------------
// update Speed via tick
// boolean型はtrue/falseの２値を表現する型
//--------------------------------------------------------------------------------------------------
boolean isTime(unsigned long *timeMark, unsigned long timeInterval) {
  unsigned long timeNow = millis();
  if ( timeNow - *timeMark >= timeInterval) {
    *timeMark = timeNow;
    return true;
  }
  return false;
}

/***********************************************************************
  OLEDの関数
************************************************************************/
void printString(const char *s, int x, int y)
{
  LCD.printString(s, x, y);
}

void printStringNOT(const char *s, int x, int y)
{
  LCD.printStringNOT(s, x, y);
}

void printStringBig(const char *s, int x, int y)
{
  LCD.printBigNumberX2(s, x, y);
}

void drawIcon(byte x, byte y, byte num, byte inNot)
{
  LCD.drawIcon(x, y, num, inNot);
}

void drawSpeedBar(byte x, byte y, byte speed, byte inNot)
{
  LCD.drawSpeedBar(x, y, speed, false);
}

void clearDisplay()
{
  LCD.lcd_clear();           //clear the screen and set start position to top left corner
}
void outputDammyPulse()
{
}
/*************************************************************************************************
  USB　Host Sheildの受信イベント
 *************************************************************************************************/
void MyJoystickEvents::OnGamePadChanged(const GamePadEventData *evt) {
/*
          Serial.print("X1: ");
          Serial.print(evt->X , HEX);
          Serial.print("\tY1: ");
          Serial.print(evt->Y , HEX);
          Serial.print("\tX2: ");
          Serial.print(evt->Z1 , HEX);
          Serial.print("\tY2: ");
          Serial.print(evt->Z2 , HEX);
          Serial.print("\tRz: ");
          Serial.print(evt->Rz , BIN);
          Serial.println("");
*/  
  buttonBufRz = evt->Rz ;
  buttonBufX1 = evt->X ;
  buttonBufX2 = evt->Z1 ;
  buttonBufY1 = evt->Y;

  //マスコンハンドルの文字表示
  switch (buttonBufRz)
  {
    case 0:
      sprintf(gText, "EB");
      LCD.printStringNOT(gText, 14, 0);
      break;
    case 0x05:
      sprintf(gText, "B8");
      LCD.printString(gText, 14, 0);
      break;
    case 0x13:
      sprintf(gText, "B7");
      LCD.printString(gText, 14, 0);
      break;
    case 0x20:
      sprintf(gText, "B6");
      LCD.printString(gText, 14, 0);
      break;
    case 0x2E:
      sprintf(gText, "B5");
      LCD.printString(gText, 14, 0);
      break;
    case 0x3C:
      sprintf(gText, "B4");
      LCD.printString(gText, 14, 0);
      break;
    case 0x49:
      sprintf(gText, "B3");
      LCD.printString(gText, 14, 0);
      break;
    case 0x57:
      sprintf(gText, "B2");
      LCD.printString(gText, 14, 0);
      break;
    case 0x65:
      sprintf(gText, "B1");
      LCD.printString(gText, 14, 0);
      break;
    case 0x80:
      sprintf(gText, "N ");
      LCD.printString(gText, 14, 0);
      break;
    case 0x9F:
      sprintf(gText, "P1");
      LCD.printString(gText, 14, 0);
      break;
    case 0xB7:
      sprintf(gText, "P2");
      LCD.printString(gText, 14, 0);
      break;
    case 0xCE:
      sprintf(gText, "P3");
      LCD.printString(gText, 14, 0);
      break;
    case 0xE6:
      sprintf(gText, "P4");
      LCD.printString(gText, 14, 0);
      break;
    case 0xFF:
      sprintf(gText, "P5");
      LCD.printString(gText, 14, 0);
      break;
  }
  /********************************************************
     以下このイベント内で使用しないのでコメントアウトにする
   ********************************************************/
  /*
          //マスコンボタン処理その１
          if ((buttonBufX1 & 0x04) == 0x04 ) {         //ボタンA
          }
          if ((buttonBufX1 & 0x02) == 0x02 ) {         //ボタンB
          }
          if ((buttonBufX1 & 0x01) == 0x01 ) {         //ボタンY
          }
          if ((buttonBufX1 & 0x08) == 0x08 ) {         //ボタンX
          }
          if ((buttonBufX1 & 0x20) == 0x20 ) {         //ボタンR
          }
          if ((buttonBufX1 & 0x10) == 0x10 ) {         //ボタンL
          }
          if ((buttonBufX1 & 0x40) == 0x40 ) {         //ボタンZL
          }
          if ((buttonBufX1 & 0x80) == 0x80 ) {         //ボタンZR
          }
          //マスコンボタン処理その２（矢印キー）
          switch(buttonBufX2)
          {
           case 0x00:        //ボタン↑
              break;
           case 0x02:        //ボタン→
              break;
           case 0x04:        //ボタン↓
              break;
           case 0x06:        //ボタン←
              break;
           //default:
          }
  */

}

//以下はLocoNetでのコールバック関数 Reference AYA Loco DCC0503
//--------------------------------------------------------------------------------------------------
// Throttle notify Call-back functions
//--------------------------------------------------------------------------------------------------
void notifyThrottleSpeed( uint8_t UserData, TH_STATE State, uint8_t Speed )
{
 // Serial.print("notifyThrottleSpeed():");
 // Serial.println(Speed);
}

//--------------------------------------------------------------------------------------------------
// Addressは、車両アドレス Slotはコマンドステーションで車両アドレスが格納されている番号
//--------------------------------------------------------------------------------------------------
void notifyThrottleAddress( uint8_t UserData, TH_STATE State, uint16_t Address, uint8_t isSlot )
{
  /*  Serial.print("notifyThrottleAddress():adress");
    Serial.print(Address);
    Serial.print("_slot:");
    Serial.println(isSlot);
  */
  SlotNo = isSlot ;
}

//--------------------------------------------------------------------------------------------------
// Notification when direction is set
//--------------------------------------------------------------------------------------------------
void notifyThrottleDirection( uint8_t UserData, TH_STATE State, uint8_t Direction )
{
  //Serial.print("notifyThrottleDirection():");
  //Serial.println(Direction);
}

//--------------------------------------------------------------------------------------------------
//Function settings.
//This library supports up to F8.
//The specifications of the F9-F28 function are not disclosed because it is OPC_IMM_PACKET..
//--------------------------------------------------------------------------------------------------
void notifyThrottleFunction( uint8_t UserData, uint8_t Function, uint8_t Value ) {
//    Serial.print("notifyThrottleFunction():");
//    Serial.print("F");
//    Serial.print(Function);
//    Serial.print(":");
//    Serial.println(Value);
      ValueFunction[Function] = Value ;   //これは消さないで　アドレス取得時にファクションの状態が拾えません。
    //  return;

}

//--------------------------------------------------------------------------------------------------
//Expected to return IDOL or Busy
//--------------------------------------------------------------------------------------------------
void notifyThrottleSlotStatus( uint8_t UserData, uint8_t Status ) {
//Serial.print("notifyThrottleSlotStatus:");
//Serial.println(Status);
}

void notifyThrottleState( uint8_t UserData, TH_STATE PrevState, TH_STATE State ) {
//    Serial.print("ThrottleState:");
//    Serial.println(State);
//    Serial.print("notifyThrottleState():");
//    Serial.println(Throttle.getStateStr(State));
}

//--------------------------------------------------------------------------------------------------
//スロットルの設定状態を返す
//--------------------------------------------------------------------------------------------------
void notifyThrottleError( uint8_t UserData, TH_ERROR Error ) {
//    Serial.print("notifyThrottleError():");
//    Serial.println(Throttle.getErrorStr(Error));
      gThotteError = Error ;            //ここでアドレス取得時の状態でスロットがInUSEか判断するために設定しています
}

//--------------------------------------------------------------------------------------------------
//Notifications for accessory control requests
//--------------------------------------------------------------------------------------------------
void notifySwitchRequest( uint16_t Address, uint8_t outData, uint8_t dirData) {
//  Serial.print("notifySwitchRequest():");
//  Serial.print(Address,DEC);
//  Serial.print(",");
//  Serial.print(outData,DEC); 
//  Serial.print(",");
//  Serial.print(dirData,DEC); // 0:t:分岐  32:c:直線
//  Serial.println("");
//  PointDir = dirData;
}

//--------------------------------------------------------------------------------------------------
